@@ -1,3 +1,11 @@
+/*
+* Copyright (c) 2013-2020, The PurpleI2P Project
+*
+* This file is part of Purple i2pd project and licensed under BSD3
+*
+* See full license text in LICENSE file at top of project tree
+*/
+
 #ifndef TRANSPORTS_H__
 #define TRANSPORTS_H__
 
@@ -15,6 +23,7 @@
 #include "TransportSession.h"
 #include "NTCPSession.h"
 #include "SSU.h"
+#include "NTCP2.h"
 #include "RouterInfo.h"
 #include "I2NPProtocol.h"
 #include "Identity.h"
@@ -80,6 +89,7 @@ namespace transport
 
 			bool IsBoundNTCP() const { return m_NTCPServer != nullptr; }
 			bool IsBoundSSU() const { return m_SSUServer != nullptr; }
+			bool IsBoundNTCP2() const { return m_NTCP2Server != nullptr; }
 
 			bool IsOnline() const { return m_IsOnline; };
 			void SetOnline (bool online) { m_IsOnline = online; };
@@ -90,7 +100,6 @@ namespace transport
 
 			void SendMessage (const i2p::data::IdentHash& ident, std::shared_ptr<i2p::I2NPMessage> msg);
 			void SendMessages (const i2p::data::IdentHash& ident, const std::vector<std::shared_ptr<i2p::I2NPMessage> >& msgs);
-			void CloseSession (std::shared_ptr<const i2p::data::RouterInfo> router);
 
 			void PeerConnected (std::shared_ptr<TransportSession> session);
 			void PeerDisconnected (std::shared_ptr<TransportSession> session);
@@ -102,7 +111,7 @@ namespace transport
 			uint64_t GetTotalReceivedBytes () const { return m_TotalReceivedBytes; };
 			uint64_t GetTotalTransitTransmittedBytes () const { return m_TotalTransitTransmittedBytes; }
 			void UpdateTotalTransitTransmittedBytes (uint32_t add) { m_TotalTransitTransmittedBytes += add; };
-			uint32_t GetInBandwidth  () const { return m_InBandwidth; };
+			uint32_t GetInBandwidth () const { return m_InBandwidth; };
 			uint32_t GetOutBandwidth () const { return m_OutBandwidth; };
 			uint32_t GetTransitBandwidth () const { return m_TransitBandwidth; };
 			bool IsBandwidthExceeded () const;
@@ -110,16 +119,16 @@ namespace transport
 			size_t GetNumPeers () const { return m_Peers.size (); };
 			std::shared_ptr<const i2p::data::RouterInfo> GetRandomPeer () const;
 
-    /** get a trusted first hop for restricted routes */
-    std::shared_ptr<const i2p::data::RouterInfo> GetRestrictedPeer() const;
-    /** do we want to use restricted routes? */
-    bool RoutesRestricted() const;
-    /** restrict routes to use only these router families for first hops */
-    void RestrictRoutesToFamilies(std::set<std::string> families);
-    /** restrict routes to use only these routers for first hops */
-    void RestrictRoutesToRouters(std::set<i2p::data::IdentHash> routers);
+			/** get a trusted first hop for restricted routes */
+			std::shared_ptr<const i2p::data::RouterInfo> GetRestrictedPeer() const;
+			/** do we want to use restricted routes? */
+			bool RoutesRestricted() const;
+			/** restrict routes to use only these router families for first hops */
+			void RestrictRoutesToFamilies(std::set<std::string> families);
+			/** restrict routes to use only these routers for first hops */
+			void RestrictRoutesToRouters(std::set<i2p::data::IdentHash> routers);
 
-    bool IsRestrictedPeer(const i2p::data::IdentHash & ident) const;
+			bool IsRestrictedPeer(const i2p::data::IdentHash & ident) const;
 
 			void PeerTest ();
 
@@ -129,18 +138,10 @@ namespace transport
 			void RequestComplete (std::shared_ptr<const i2p::data::RouterInfo> r, const i2p::data::IdentHash& ident);
 			void HandleRequestComplete (std::shared_ptr<const i2p::data::RouterInfo> r, i2p::data::IdentHash ident);
 			void PostMessages (i2p::data::IdentHash ident, std::vector<std::shared_ptr<i2p::I2NPMessage> > msgs);
-			void PostCloseSession (std::shared_ptr<const i2p::data::RouterInfo> router);
 			bool ConnectToPeer (const i2p::data::IdentHash& ident, Peer& peer);
 			void HandlePeerCleanupTimer (const boost::system::error_code& ecode);
 			void HandlePeerTestTimer (const boost::system::error_code& ecode);
 
-			void NTCPResolve (const std::string& addr, const i2p::data::IdentHash& ident);
-			void HandleNTCPResolve (const boost::system::error_code& ecode, boost::asio::ip::tcp::resolver::iterator it,
-				i2p::data::IdentHash ident, std::shared_ptr<boost::asio::ip::tcp::resolver> resolver);
-			void SSUResolve (const std::string& addr, const i2p::data::IdentHash& ident);
-			void HandleSSUResolve (const boost::system::error_code& ecode, boost::asio::ip::tcp::resolver::iterator it,
-				i2p::data::IdentHash ident, std::shared_ptr<boost::asio::ip::tcp::resolver> resolver);
-
 			void UpdateBandwidth ();
 			void DetectExternalIP ();
 
@@ -154,6 +155,7 @@ namespace transport
 
 			NTCPServer * m_NTCPServer;
 			SSUServer * m_SSUServer;
+			NTCP2Server * m_NTCP2Server;
 			mutable std::mutex m_PeersMutex;
 			std::map<i2p::data::IdentHash, Peer> m_Peers;
 
@@ -179,6 +181,7 @@ namespace transport
 			// for HTTP only
 			const NTCPServer * GetNTCPServer () const { return m_NTCPServer; };
 			const SSUServer * GetSSUServer () const { return m_SSUServer; };
+			const NTCP2Server * GetNTCP2Server () const { return m_NTCP2Server; };
 			const decltype(m_Peers)& GetPeers () const { return m_Peers; };
 	};
 
