@@ -1,4 +1,13 @@
+/*
+* Copyright (c) 2013-2020, The PurpleI2P Project
+*
+* This file is part of Purple i2pd project and licensed under BSD3
+*
+* See full license text in LICENSE file at top of project tree
+*/
+
 #include <algorithm>
+#include <random>
 #include "I2PEndian.h"
 #include "Crypto.h"
 #include "Tunnel.h"
@@ -10,15 +19,11 @@
 #include "Tunnel.h"
 #include "TunnelPool.h"
 #include "Destination.h"
-#ifdef WITH_EVENTS
-#include "Event.h"
-#endif
 
 namespace i2p
 {
 namespace tunnel
 {
-
 	TunnelPool::TunnelPool (int numInboundHops, int numOutboundHops, int numInboundTunnels, int numOutboundTunnels):
 		m_NumInboundHops (numInboundHops), m_NumOutboundHops (numOutboundHops),
 		m_NumInboundTunnels (numInboundTunnels), m_NumOutboundTunnels (numOutboundTunnels), m_IsActive (true),
@@ -69,7 +74,8 @@ namespace tunnel
 		m_Tests.clear ();
 	}
 
-	bool TunnelPool::Reconfigure(int inHops, int outHops, int inQuant, int outQuant)	{
+	bool TunnelPool::Reconfigure(int inHops, int outHops, int inQuant, int outQuant)
+	{
 		if( inHops >= 0 && outHops >= 0 && inQuant > 0 && outQuant > 0)
 		{
 			m_NumInboundHops = inHops;
@@ -80,30 +86,22 @@ namespace tunnel
 		}
 		return false;
 	}
-	
+
 	void TunnelPool::TunnelCreated (std::shared_ptr<InboundTunnel> createdTunnel)
 	{
 		if (!m_IsActive) return;
 		{
-#ifdef WITH_EVENTS
-			EmitTunnelEvent("tunnels.created", createdTunnel);
-#endif
 			std::unique_lock<std::mutex> l(m_InboundTunnelsMutex);
 			m_InboundTunnels.insert (createdTunnel);
 		}
 		if (m_LocalDestination)
 			m_LocalDestination->SetLeaseSetUpdated ();
-
-		OnTunnelBuildResult(createdTunnel, eBuildResultOkay);
 	}
 
 	void TunnelPool::TunnelExpired (std::shared_ptr<InboundTunnel> expiredTunnel)
 	{
 		if (expiredTunnel)
 		{
-#ifdef WITH_EVENTS
-			EmitTunnelEvent("tunnels.expired", expiredTunnel);
-#endif
 			expiredTunnel->SetTunnelPool (nullptr);
 			for (auto& it: m_Tests)
 				if (it.second.second == expiredTunnel) it.second.second = nullptr;
@@ -117,14 +115,9 @@ namespace tunnel
 	{
 		if (!m_IsActive) return;
 		{
-#ifdef WITH_EVENTS
-			EmitTunnelEvent("tunnels.created", createdTunnel);
-#endif
 			std::unique_lock<std::mutex> l(m_OutboundTunnelsMutex);
 			m_OutboundTunnels.insert (createdTunnel);
 		}
-		OnTunnelBuildResult(createdTunnel, eBuildResultOkay);
-
 		//CreatePairedInboundTunnel (createdTunnel);
 	}
 
@@ -132,9 +125,6 @@ namespace tunnel
 	{
 		if (expiredTunnel)
 		{
-#ifdef WITH_EVENTS
-			EmitTunnelEvent("tunnels.expired", expiredTunnel);
-#endif
 			expiredTunnel->SetTunnelPool (nullptr);
 			for (auto& it: m_Tests)
 				if (it.second.first == expiredTunnel) it.second.first = nullptr;
@@ -198,8 +188,8 @@ namespace tunnel
 			{
 				if (it->IsEstablished () && it != excluded)
 				{
-						tunnel = it;
-						i++;
+					tunnel = it;
+					i++;
 				}
 				if (i > ind && tunnel) break;
 			}
@@ -429,7 +419,7 @@ namespace tunnel
 		{
 			std::lock_guard<std::mutex> lock(m_CustomPeerSelectorMutex);
 			if (m_CustomPeerSelector)
-					return m_CustomPeerSelector->SelectPeers(peers, numHops, isInbound);
+				return m_CustomPeerSelector->SelectPeers(peers, numHops, isInbound);
 		}
 		// explicit peers in use
 		if (m_ExplicitPeers) return SelectExplicitPeers (peers, isInbound);
@@ -441,7 +431,7 @@ namespace tunnel
 		int size = m_ExplicitPeers->size ();
 		std::vector<int> peerIndicies;
 		for (int i = 0; i < size; i++) peerIndicies.push_back(i);
-		std::random_shuffle (peerIndicies.begin(), peerIndicies.end());
+		std::shuffle (peerIndicies.begin(), peerIndicies.end(), std::mt19937(std::random_device()()));
 
 		int numHops = isInbound ? m_NumInboundHops : m_NumOutboundHops;
 		for (int i = 0; i < numHops; i++)
@@ -610,11 +600,5 @@ namespace tunnel
 		}
 		return tun;
 	}
-
-	void TunnelPool::OnTunnelBuildResult(std::shared_ptr<Tunnel> tunnel, TunnelBuildResult result)
-	{
-		auto peers = tunnel->GetPeers();
-		if(m_CustomPeerSelector) m_CustomPeerSelector->OnBuildResult(peers, tunnel->IsInbound(), result);
-	}
 }
 }
